#![allow(unused)]

use errno;
use std;

/// General error type.
#[derive(Clone, Debug)]
pub enum Error {
    /// Error originating from Rust code.
    InternalError { description: String, cause: String },
    /// Error originating from FFI calls.
    ForeignError {
        api_name: &'static str,
        errno: errno::Errno,
    },
}

impl Error {
    pub(crate) fn foreign_with_errno(api_name: &'static str) -> Error {
        Error::ForeignError {
            api_name,
            errno: errno::errno(),
        }
    }
    pub(crate) fn internal_with_cause<E>(description: String, cause: &E) -> Error
    where
        E: std::error::Error,
    {
        Error::InternalError {
            description,
            cause: String::from(cause.description()),
        }
    }
}

impl std::fmt::Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
        match self {
            Error::InternalError { description, cause } => {
                write!(f, "InternalError: {} ({})", description, cause)
            }
            Error::ForeignError { api_name, errno } => write!(
                f,
                "ForeignError with code {} occured in `{}`: {}",
                errno.0, api_name, errno
            ),
        }
    }
}

impl std::error::Error for Error {}
